/*
  MegaTool V1.25 by vOlker Oth (dOnut) in 7/1997
  based on MeTool 1.85
  For bug reports and comments mail to
    volkeroth@aol.com
  You might find a more advanced version someday at my homepage:
    members.aol.com/volkeroth

  You need a properly installed Allegro 2.2 to compile MegaTool.
  To avoid crashes, insert the following line at the end of
  allegro\src\djgpp\mouse.c:
  _mouse_screen = NULL;
*/

#include <stdio.h>
#include <crt0.h>
#include <string.h>
#include <dir.h>
#include <unistd.h>
#include <bios.h>
#include <allegro.h>
#include <process.h>
#include "guidat.h"

/*constants*/
#define  true -1
#define  false 0
#define  error 1

#define _SMD   1                                    /* different rom types */
#define _BIN   2                               /* used in file_info[]->rom */
#define _SPLIT 3

#define _modex      8                 /* values for internal option format */
#define _nojoy     16
#define _nosound   32
#define _z80       64
#define _fakedz80 128

#define w_all 1                       /* values for state of radio buttons */
#define w_rom 2

#define  maxfiles 1024
#define  maxlines 1024
#define  maxoptions 1024
#define  FA_MYFILE FA_RDONLY + FA_HIDDEN + FA_SYSTEM + FA_ARCH

#define  ident_exec       "[Exec]"                /* some string constants */
#define  ident_startpath  "[StartPath]"
#define  ident_romsonly   "[RomsOnly]"
#define  ident_g_options  "[GlobalOptions]"
#define  ident_l_options  "[LocalOptions]"
#define  io_err           "I/O Error"
#define  io_err_inf       "<i/o error>"
#define  mem_err          "Memory Error"
#define  mem_err_nl       "Memory Error\n"
#define  wild_all         "*.*"

/* macros */
#define debug(c) puts(c)
#define swap_hilo(i) ((i & 0xff) <<8) | ((i & 0xff00) >> 8)
#define swap_dword(i) ((i & 0xff)<< 24)|((i & 0xff00)<<8)\
                     |((i & 0xff0000)>>8)|((i & 0xff000000)>>24)
/*typedefs*/
typedef unsigned char byte;
typedef unsigned short int word;
typedef unsigned long int dword;

typedef struct {                          /* the embedded header structure */
  char segatitle[16];                                 /* 'SEGA MEGA DRIVE' */
  char copyright[16];                                 /* '(C)SEGA 1988.JUL */
  char domestic[48];                                  /* japanese name     */
  char overseas[48];                                  /* us/european name  */ 
  char production[14];                                /* production code   */
  word checksum;                                      /* rom check sum     */
  char control[16];                                   /* control data      */
  dword rom_start;
  dword rom_end;
  dword ram_start;
  dword ram_end;
  char extram[12];                                    /* external ram info */
  char modem[12];                                     /* modem info        */
  char memo[40];                                      /* MEMO ???          */
  char country[16];                                   /* country code      */
} t_header;

typedef struct {                        /* used to store file informations */
  char name[13];
  long size;                                                  /* file size */
  byte dir;                              /* boolean: is this a directory ? */
  byte rom;                         /* boolean: is there a rom extension ? */
} t_file_info;

typedef byte block16k[16384];            /* used for loading/saving blocks */

typedef struct {                                  /* used to store options */
  char name[9];
  word option;
} t_local_option;

/*  dialog definitions */
enum my_dialog_ix {
  #define DIA_ENTRY( proc, mnem, x, y, w, h, fg, bg, key, flags, d1, d2, dp )\
      mnem
  #include "megatool.dia"
};

enum filereq_ix {
  #define DIA_ENTRY( proc, mnem, x, y, w, h, fg, bg, key, flags, d1, d2, dp )\
      mnem
  #include "filereq.dia"
};

enum setup_ix {
  #define DIA_ENTRY( proc, mnem, x, y, w, h, fg, bg, key, flags, d1, d2, dp )\
      mnem
  #include "setup.dia"
};

enum lsetup_ix {
  #define DIA_ENTRY( proc, mnem, x, y, w, h, fg, bg, key, flags, d1, d2, dp )\
      mnem
  #include "lsetup.dia"
};

enum rename_ix {
  #define DIA_ENTRY( proc, mnem, x, y, w, h, fg, bg, key, flags, d1, d2, dp )\
      mnem
  #include "rename.dia"
};

/* variables */

char* options[] = { 0,0,0,0,0,0,0 };                      /* used in spawn */
t_header* geheader;                          /* pointer to embedded header */
block16k header_buf;       /* used for reading header/writing header block */
char* text_list[maxlines];                      /* used for sorting output */
int line_num = 0;                          /* number of lines in text_list */
t_local_option* local_option[maxoptions];   /* used to store local options */
int opt_num;                                   /* number of option entries */
int opt_idx;                               /* index of actual option entry */
t_file_info* file_info[maxfiles];                     /* contains fileinfo */
int file_num = 0;                    /* the actual number of fitting files */
int file_idx = 0;                      /* index of actual file in file_num */
int old_file_idx=0;          /* used to store idx when creating list files */
FILE *f;                                           /* standard file handle */
struct ffblk dirinfo;                          /* needed for dir functions */
int doserror;                         /* return variable for file-functions*/

char str_ini[128];                              /* holds path of ini files */
char str_fname[13];                     /* string holding current filename */
char str_rename[13];                     /* string holding rename filename */
char str_path[128];                         /* string holding current path */
char str_startpath[128]=".";                                 /* start path */
char str_genem[128]="genem.exe";                        /* genem exec path */
char str_view[13];                                  /* used in View button */
char str_exec[128]="\0";                          /* used to execute genem */
char str_wild[13];                          /* wild card used in get_files */
char str_info[80] = "<no rom>";         /* string holding current rom info */
char str_blocks[80]= "0";              /* string holding current block no. */
char str_frame[2];                          /* actual frame setting string */
char str_debug[18];
char str_name[15];

char str_overseas[49]="\0";            /* string holding current ov. title */
char str_domestic[49]="\0";           /* string holding current dom. title */
char str_copyright[17]="\0";        /* string holding current copyr. info  */
char str_production[15]="\0";    /* string holding current production info */
char str_checksum[15]="\0";             /* string holding current checksum */
char str_control[17]="\0";          /* string holding current control info */
char str_country[17]="\0";          /* string holding current country info */

char fr_path[128];                          /* path used in file requester */
char fr_fname[13];                          /* name used in file requester */
char fr_title[50];                         /* title used in file requester */

int roms_only=false;                                     /* show only roms */
int smd = false;                               /* boolean: is this a smd ? */
int bin = false;                               /* boolean: is this a bin ? */
int split = false;                       /* boolean: is this a split smd ? */
int japan=false;
int usa=false;
int europe=false;
int changed=false;                           /* boolean: was rom changed ? */
int leave=false;                               /* boolean: leave program ? */
int wild=w_all;
int blocks=0;                                 /* current 16kb block number */
int rblocks=0;                                 /* block reported in header */
word checksum =0;                                      /* current checksum */
word g_option=2;                                         /* global options */
word old_gmode;
int i;                                                     /* loop counter */
int ret;                                         /* dialog return variable */


/* Misc functions */

char** __crt0_glob_function(char *_argument) { return(NULL);}

void __crt0_load_environment_file(char *_appname) {}

/* GUI keyboard functions */

int replacement_keypressed()
{
  return _bios_keybrd(_KEYBRD_READY);
}

int replacement_readkey()
{
  return _bios_keybrd(_KEYBRD_READ);
}

/* GUI functions */

static int gui_textout(BITMAP *bmp, char *s, int x, int y, int color, int centre)
/* copied from allegro's gui.c since it is not exported */
{
  char tmp[160];
  int c;
  int len = 0;
  int pix_len;
  int hline_pos = -1;

  for (c=0; (s[c]) && (len<159); c++) {
    if (s[c] == '&') {
	   if (s[c+1] != '&') hline_pos = len;
	   else {
        tmp[len++] = '&';
	     c++;
      }
    }
    else tmp[len++] = s[c];
  }
  tmp[len] = 0;
  pix_len = text_length(font, tmp);
  if (centre) x -= pix_len / 2;
  if (bmp) {
    textout(bmp, font, tmp, x, y, color);
    if (hline_pos >= 0) {
	   c = tmp[hline_pos];
	   tmp[hline_pos] = 0;
	   hline_pos = text_length(font, tmp);
	   tmp[0] = c;
	   tmp[1] = 0;
	   c = text_length(font, tmp);
	   hline(bmp, x+hline_pos, y+text_height(font)-gui_font_baseline, x+hline_pos+c-1, color);
    }
  }
  return pix_len;
}

static int gui_strlen(char *s)
/* copied from allegro's gui.c since it is not exported */
{
   return gui_textout(NULL, s, 0, 0, 0, 0);
}

static void dotted_rect(int x1, int y1, int x2, int y2, int fg, int bg)
/* copied from allegro's gui.c since it is not exported */
/* Draws a dotted rectangle, for showing an object has the input focus.*/
{
  int c;
  for (c=x1; c<=x2; c+=2) {                /* "<=" else a pixel is missing */
    putpixel(screen, c, y1, fg);
    putpixel(screen, c, y2, fg);
  }
  for (c=x1+1; c<x2; c+=2) {
    putpixel(screen, c, y1, bg);
    putpixel(screen, c, y2, bg);
  }
  for (c=y1; c<y2; c+=2) {
    putpixel(screen, x1, c, fg);
    putpixel(screen, x2, c, fg);
  }
  for (c=y1+1; c<y2; c+=2) {
    putpixel(screen, x1, c, bg);
    putpixel(screen, x2, c, bg);
  }
}

int d_3d_box_proc(int msg, DIALOG *d, int c)
/* Modified d_box_proc. draws "3d" box */
{
  int col1=6;
  int col2=8;

  if (msg==MSG_DRAW && !(d->flags & D_HIDDEN)) {
    if (!(d->flags & D_DISABLED)) {
      rectfill(screen, d->x+2, d->y+2, d->x+d->w-2, d->y+d->h-2, 7);
      rect(screen, d->x, d->y, d->x+d->w, d->y+d->h, 1);
      col1=8; col2 = 6;
    }
    vline(screen, d->x+1, d->y+1, d->y+d->h-1, col1);
    hline(screen, d->x+1, d->y+1, d->x+d->w-1, col1);
    vline(screen, d->x+d->w-1, d->y+2, d->y+d->h-1, col2);
    hline(screen, d->x+2, d->y+d->h-1, d->x+d->w-1, col2);
  }
  return D_O_K;
}

int d_3d_button_proc(int msg, DIALOG *d, int c)
/* modified d_button_proc: uses 3d effect*/
  {
  int fg;
  int col1;
  int col2;
  int g;

  if (msg==MSG_DRAW) {
    fg = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;
    if ((d->flags & D_SELECTED) || (d->flags & D_DISABLED)) {
      col1=2; col2=3; g=1;
    }
    else {
      col1=4; col2=2; g=0;
    }
    rect(screen, d->x, d->y, d->x+d->w, d->y+d->h, 1);
    rectfill(screen, d->x+2, d->y+2, d->x+d->w-2, d->y+d->h-2, 3);
	 text_mode(-1);
	 gui_textout(screen, d->dp, d->x+d->w/2+g, d->y+d->h/2-text_height(font)/2+g, fg, TRUE);
    vline(screen, d->x+1, d->y+1, d->y+d->h-1, col1);
    hline(screen, d->x+1, d->y+1, d->x+d->w-1, col1);
    vline(screen, d->x+d->w-1, d->y+1+g, d->y+d->h-1, col2);
    hline(screen, d->x+1+g, d->y+d->h-1, d->x+d->w-1, col2);
    if (d->flags & D_GOTFOCUS)
      dotted_rect(d->x+3, d->y+3, d->x+d->w-3, d->y+d->h-3, fg, d->bg);
    return D_O_K;
  }
  return d_button_proc(msg, d, 0);
}

int d_3d_check_proc(int msg, DIALOG *d, int c)
/* modified d_check_proc */
{
  int x;
  int fg;
  BITMAP *spr_check;

  if (msg==MSG_DRAW) {
    fg = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;
    text_mode(-1);
    x = d->x + gui_textout(screen, d->dp, d->x, d->y+(d->h-(text_height(font)-gui_font_baseline))/2, fg, FALSE) + text_height(font)/2;
    if (d->flags & D_SELECTED) {
      if (d->flags & D_DISABLED)
        spr_check = &spr_check1_p;
      else
        spr_check = &spr_check1;
    }
    else {
      if (d->flags & D_DISABLED)
        spr_check = &spr_check2_p;
      else
         spr_check = &spr_check2;
    }
    draw_sprite(screen,spr_check,x,d->y);
    if (d->flags & D_GOTFOCUS)
      dotted_rect(x, d->y, x+d->h-2, d->y+d->h-2, fg, d->bg);
    return D_O_K;
  }
  return d_button_proc(msg, d, 0);
}

int d_3d_radio_proc(int msg, DIALOG *d, int c)
/* radio button based on d_button_proc */
/* the radio button behaviour is controlled by the program */
/* in opposite to the standard d_radio_proc of the gui */
{
  int x, center, fg;

  if (msg==MSG_DRAW) {
    fg = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;
    text_mode(d->bg);
    gui_textout(screen, d->dp, d->x+d->h+text_height(font), d->y+(d->h-(text_height(font)-gui_font_baseline))/2, fg, FALSE);
    x = d->x;
    if (d->flags & D_SELECTED)
      draw_sprite(screen,&spr_radio2,x,d->y);
    else
      draw_sprite(screen,&spr_radio1,x,d->y);
    if (d->flags & D_GOTFOCUS)
      dotted_rect(d->x, d->y, d->x+13, d->y+13, fg, d->bg);
    return D_O_K;
  }
  return d_button_proc(msg, d, 0);
}

static DIALOG alert3d_dialog[] =
{
   /* (dialog proc)     (x)   (y)   (w)   (h)   (fg)  (bg)  (key) (flags)  (d1)  (d2)  (dp) */
   { d_3d_box_proc,     0,    0,    0,    0,    0,    0,    0,    0,       0,    0,    NULL },
   { d_ctext_proc,      0,    0,    0,    0,    0,    0,    0,    0,       0,    0,    NULL },
   { d_ctext_proc,      0,    0,    0,    0,    0,    0,    0,    0,       0,    0,    NULL },
   { d_ctext_proc,      0,    0,    0,    0,    0,    0,    0,    0,       0,    0,    NULL },
   { d_3d_box_proc,     0,    0,    0,    0,    0,    0,    0,    0,       0,    0,    NULL },
   { d_3d_button_proc,  0,    0,    0,    0,    0,    0,    0,    D_EXIT,  0,    0,    NULL },
   { d_3d_box_proc,     0,    0,    0,    0,    0,    0,    0,    0,       0,    0,    NULL },
   { d_3d_button_proc,  0,    0,    0,    0,    0,    0,    0,    D_EXIT,  0,    0,    NULL },
   { d_3d_box_proc,     0,    0,    0,    0,    0,    0,    0,    0,       0,    0,    NULL },
   { d_3d_button_proc,  0,    0,    0,    0,    0,    0,    0,    D_EXIT,  0,    0,    NULL },
   { NULL }
};

#define A_S1  1
#define A_S2  2
#define A_S3  3
#define A_R1  4
#define A_B1  5
#define A_R2  6
#define A_B2  7
#define A_R3  8
#define A_B3  9

int alert3d(char *s1, char *s2, char *s3, char *b1, char *b2, char *b3, int c1, int c2, int c3)
/* modified alert3 to get 3d effect */
  {
  int maxlen = 0;
  int len1, len2, len3;
  int avg_w = text_length(font, " ");
  int avg_h = text_height(font)+2;             /* +2 to make it look nicer */
  int avg_hx = avg_h+3;                        /* +5 to make it look nicer */
  int buttons = 0;
  int b[3];
  int c;

#define SORT_OUT_BUTTON(x) {                                           \
  if (b##x) {                                                          \
    alert3d_dialog[A_B##x].flags &= ~D_HIDDEN;                         \
    alert3d_dialog[A_B##x].key = c##x;                                 \
    alert3d_dialog[A_B##x].dp = b##x;                                  \
    len##x = gui_strlen(b##x);                                         \
	 b[buttons++] = A_B##x;                                             \
  }                                                                    \
  else {                                                               \
  	 alert3d_dialog[A_B##x].flags |= D_HIDDEN;                          \
	 len##x = 0;                                                        \
  }                                                                    \
}
  alert3d_dialog[A_S1].dp = alert3d_dialog[A_S2].dp = alert3d_dialog[A_S3].dp =
		   alert3d_dialog[A_B1].dp = alert3d_dialog[A_B2].dp = "";
  if (s1) {
    alert3d_dialog[A_S1].dp = s1;
    maxlen = text_length(font, s1);
  }
  if (s2) {
    alert3d_dialog[A_S2].dp = s2;
    len1 = text_length(font, s2);
    if (len1 > maxlen) maxlen = len1;
  }
  if (s3) {
    alert3d_dialog[A_S3].dp = s3;
    len1 = text_length(font, s3);
    if (len1 > maxlen) maxlen = len1;
  }
  SORT_OUT_BUTTON(1);
  SORT_OUT_BUTTON(2);
  SORT_OUT_BUTTON(3);
  len1 = MAX(len1, MAX(len2, len3)) + avg_w*3;
  if (len1*buttons > maxlen) maxlen = len1*buttons;
  maxlen += avg_w*4;
  alert3d_dialog[0].w = maxlen;
  alert3d_dialog[A_S1].x = alert3d_dialog[A_S2].x = alert3d_dialog[A_S3].x =
					alert3d_dialog[0].x + maxlen/2;
  alert3d_dialog[A_B1].w = alert3d_dialog[A_B2].w = alert3d_dialog[A_B3].w = len1;
  alert3d_dialog[A_B1].x = alert3d_dialog[A_B2].x = alert3d_dialog[A_B3].x =
		         alert3d_dialog[0].x + maxlen/2 - len1/2;
  if (buttons == 3) {
    alert3d_dialog[b[0]].x = alert3d_dialog[0].x + maxlen/2 - len1*3/2 - avg_w;
    alert3d_dialog[b[2]].x = alert3d_dialog[0].x + maxlen/2 + len1/2 + avg_w;
  }
  else if (buttons == 2) {
    alert3d_dialog[b[0]].x = alert3d_dialog[0].x + maxlen/2 - len1 - avg_w;
    alert3d_dialog[b[1]].x = alert3d_dialog[0].x + maxlen/2 + avg_w;
  }
  alert3d_dialog[0].h = avg_hx*7;                               /* changed */
  alert3d_dialog[A_S1].y = alert3d_dialog[0].y + avg_hx;        /* changed */
  alert3d_dialog[A_S2].y = alert3d_dialog[0].y + avg_hx*2;      /* changed */
  alert3d_dialog[A_S3].y = alert3d_dialog[0].y + avg_hx*3;      /* changed */
  alert3d_dialog[A_S1].h = alert3d_dialog[A_S2].h = alert3d_dialog[A_S2].h = avg_hx; /* changed */
  alert3d_dialog[A_B1].y = alert3d_dialog[A_B2].y = alert3d_dialog[A_B3].y =
						alert3d_dialog[0].y + avg_hx*3+avg_h*2;       /* changed */
  alert3d_dialog[A_B1].h = alert3d_dialog[A_B2].h = alert3d_dialog[A_B3].h = avg_h*2;
  alert3d_dialog[A_R1].x = alert3d_dialog[A_B1].x-2;
  alert3d_dialog[A_R1].y = alert3d_dialog[A_B1].y-2;
  alert3d_dialog[A_R1].w = alert3d_dialog[A_B1].w+4;
  alert3d_dialog[A_R1].h = alert3d_dialog[A_B1].h+4;
  alert3d_dialog[A_R1].flags = alert3d_dialog[A_B1].flags & D_HIDDEN | D_DISABLED;
  alert3d_dialog[A_R2].x = alert3d_dialog[A_B2].x-2;
  alert3d_dialog[A_R2].y = alert3d_dialog[A_B2].y-2;
  alert3d_dialog[A_R2].w = alert3d_dialog[A_B2].w+4;
  alert3d_dialog[A_R2].h = alert3d_dialog[A_B2].h+4;
  alert3d_dialog[A_R2].flags = alert3d_dialog[A_B2].flags & D_HIDDEN | D_DISABLED;
  alert3d_dialog[A_R3].x = alert3d_dialog[A_B3].x-2;
  alert3d_dialog[A_R3].y = alert3d_dialog[A_B3].y-2;
  alert3d_dialog[A_R3].w = alert3d_dialog[A_B3].w+4;
  alert3d_dialog[A_R3].h = alert3d_dialog[A_B3].h+4;
  alert3d_dialog[A_R3].flags = alert3d_dialog[A_B3].flags & D_HIDDEN |D_DISABLED;
  centre_dialog(alert3d_dialog);
  set_dialog_color(alert3d_dialog, gui_fg_color, gui_bg_color);
  clear_keybuf();
  do {} while (mouse_b);
  c = popup_dialog(alert3d_dialog, A_B1);
  if (c == A_B1) return 1;
  else if (c == A_B2) return 2;
  else return 3;
}

int d_edit_proc2(int msg, DIALOG *d, int c)
/* This is a modified d_edit_proc from allegro: */
/* It will close the dialog if return was pressed. */
/* The shown text also scrolls if there is more text */
/* than can be displayed */
{
  int f, l, p, w, x, fg;
  int b;                             /* no. of chars that can be displayed */
  int scroll=false;                 /* bool: does text have to be scrolled */
  char *s;
  char buf[2];

  s = d->dp;
  l = strlen(s);
  if (d->d2 > l) d->d2 = l;
  /* calculate maximal number of displayable characters */
  b=x=0;
  if (d->d2==l) x=text_height(font);           /* reserve place for cursor */
  buf[1] = 0;
  for (p=d->d2; p>=0; p--) {              /* step backwards through string */
    buf[0] = s[p];
    b++;
    x += text_length(font, buf);
    if (x > d->w) break;
  }
  if (x <= d->w) b=l;                     /* nope, we don't have to scroll */
  else {
    b--;                                         /* yep, we have to scroll */
    scroll=true;
  }
  switch (msg) {
    case MSG_START:
	   d->d2 = l;                     /* start with cursor at end of string */
      do {         /* find last character that doesn't have to be scrolled */
        scroll=false;
        b=x=0;
        if (d->d2==l) x=text_height(font);     /* reserve place for cursor */
        for (p=d->d2; p>=0; p--) {                /* calc max no. of chars */
          buf[0] = s[p];
          b++;
          x += text_length(font, buf);
          if (x > d->w) break;
        }
        if (x > d->w) {
          d->d2--;                                      /* loop once again */
          scroll=true;
        }
      } while (scroll);
      break;
    case MSG_DRAW:
	   fg = (d->flags & D_DISABLED) ? gui_mg_color : d->fg;
	   x = 0;
      /* handle text box scrolling */
      if (scroll) {
        p = d->d2-b+1;                          /* increase start position */
        b = d->d2;                               /* display to current pos */
      }
      else p=0;                                                /* b == l ! */
      for (p; p<=b; p++) {
	     buf[0] = s[p] ? s[p] : ' ';
	     w = text_length(font, buf);
	     if (x+w > d->w) break;
        f = ((p == d->d2) && (d->flags & D_GOTFOCUS));
        text_mode(f ? fg : d->bg);
	     textout(screen, font, buf, d->x+x, d->y, f ? d->bg : fg);
        x += w;
      }
      if (x<d->w)
        rectfill(screen, d->x+x, d->y, d->x+d->w-1, d->y+text_height(font)-1, d->bg);
      break;
    case MSG_CLICK:
	   x = d->x;
      /* handle text box scrolling */
      if (scroll) {
        p = d->d2-b+1;                          /* increase start position */
        b = d->d2;                               /* display to current pos */
      }
      else
        p=0;                                                   /* b == l ! */
      for (p; p<b; p++) {
	     buf[0] = s[p];
	     x += text_length(font, buf);
	     if (x > mouse_x) break;
      }
      d->d2 = MID(0, p, l);
	   show_mouse(NULL);
	   SEND_MESSAGE(d, MSG_DRAW, 0);
	   show_mouse(screen);
	   break;
    case MSG_WANTFOCUS:
    case MSG_LOSTFOCUS:
    case MSG_KEY: return D_WANTFOCUS;
    case MSG_CHAR:
      if ((c >> 8) == KEY_LEFT) {
        if (d->d2 > 0) d->d2--;
      }
	   else if ((c >> 8) == KEY_RIGHT) {
        if (d->d2 < l) d->d2++;
      }
	   else if ((c >> 8) == KEY_HOME) {
        d->d2 = 0;
      }
	   else if ((c >> 8) == KEY_END) {
        d->d2 = l;
      }
	   else if ((c >> 8) == KEY_DEL) {
	     if (d->d2 < l)
	       for (p=d->d2; s[p]; p++) s[p] = s[p+1];
	   }
      else if ((c >> 8) == KEY_BACKSPACE) {
	     if (d->d2 > 0) {
	       d->d2--;
	       for (p=d->d2; s[p]; p++) s[p] = s[p+1];
        }
      }
      /* close dialog if enter was pressed and D_EXIT is set */
      else if ((c >> 8) == KEY_ENTER) {
        if (d->flags & D_EXIT) {
          show_mouse(NULL);
	       SEND_MESSAGE(d, MSG_DRAW, 0);
     	    show_mouse(screen);
          return D_CLOSE;
        }
      }
	   else {
	     c &= 0xff;
	     if ((c >= 32) && (c <= 126)) {
	       if (l < d->d1) {
		      while (l >= d->d2) {
		        s[l+1] = s[l];
		        l--;
            }
            s[d->d2] = c;
		      d->d2++;
          }
        }
        else return D_O_K;
	   }
      /* if we changed something, better redraw... */
      show_mouse(NULL);
      SEND_MESSAGE(d, MSG_DRAW, 0);
      show_mouse(screen);
      return D_USED_CHAR;
  }
  return D_O_K;
}

char* get_file_list( int index, int *list_size )
/* needed by the file list box */
{
  static char str_dir[128];

  if (index < 0) {
    *list_size = file_num;
    return NULL;
  }
  else
    if (!file_info[index]->dir)
      return (char*)file_info[index]->name;
    else {
      /* show directory entrys in square brackets */
      strcpy(str_dir,"[");
      strcat(str_dir,file_info[index]->name);
      strcat(str_dir,"]");
      return (char*)str_dir;
    }
}

void ok_alert(char s1[], char s2[], char s3[])
/* shortcut for the often used 'ok' type alert */
{
  alert3d(s1,s2,s3,"&Ok",0,0,'o',0,0);
}

int full_country()
/* shows the fully country code */
/* called by a keyboard proc in the main dialog */
{
  if (smd||bin)
    ok_alert("Full Country Code:",str_country,0);
  return D_CLOSE;
}

/*  include the main dialog */
DIALOG my_dialog[] =
{
  #include "megatool.dia"
};

/* include the about dialog */
DIALOG about_dialog[] =
{
  #include "about.dia"
};

/*  include the file request dialog */
DIALOG fr_dialog[] =
{
  #include "filereq.dia"
};

/*  include the global setup dialog */
DIALOG setup_dialog[] =
{
  #include "setup.dia"
};

/*  include the local setup dialog */
DIALOG lsetup_dialog[] =
{
  #include "lsetup.dia"
};

/*  include the rename dialog */
DIALOG rename_dialog[] =
{
  #include "rename.dia"
};

void clr_scrn()
/* simply clears the screen */
{
  show_mouse(NULL);
  set_clip(screen, 0, 0, SCREEN_W-1, SCREEN_H-1);
  clear_to_color(screen, 7);
  show_mouse(screen);
}

void get_gmode()
/* gets gfx mode */
/* get_gmode() and set_gmode() are necessary since Allegro */
/* gets confused when started up more than once */
{
  __dpmi_regs r;
  r.x.ax = 0x0f00;                                          /* function 0f */
  __dpmi_int(0x10,&r);
  old_gmode = (r.x.ax & 255);                            /* gfx mode in al */
}

void set_gmode()
/* sets gfx mode */
{
  __dpmi_regs r;
  r.x.ax = old_gmode;                                        /* function 0 */
  __dpmi_int(0x10,&r);
  old_gmode = (r.x.ax & 255);                            /* gfx mode in al */
}

/* program functions */

int file_overwrite(char fname[])
/* Displays alert "file exists" */
/* returns true if file doesn't exist yet or Overwrite-Yes was selected */
{
  if (file_exists(fname, 63, NULL))
    if (alert3d("File exists !","Overwrite ?",0,"&Yes","&No",0,'y','n',0)==2)
      return false;
    else return true;
  else return true;
}

int bigger(int s1, int s2)
/* compare function used by sort_files  */
/* files are always "bigger" than directories, "z*.*" is bigger than "a*.*" */
{
  int h;
  h = file_info[s2]->dir - file_info[s1]->dir;
  if (h==0) return strcmp(file_info[s1]->name,file_info[s2]->name);
  else return h;
}

void sort_files()
/* sort files in list by name - directories before normal files */
/* used by get_files */
{
  t_file_info *hstruct;
  int i=1;
  int j=1;

  do {
    for (i= j+1; i< file_num;i++)
      if (bigger(j,i)>0) {
        hstruct = file_info[j];
        file_info[j]=file_info[i];
        file_info[i]=hstruct;
      }
    j++;
  } while (j < file_num);
}

void add_file()
/* store file informations in array file_info */
/* used by get_files */
{
  if (file_num < maxfiles)
    if (strcmp(dirinfo.ff_name,".")) {
      if (file_info[file_num] = (t_file_info*)malloc(sizeof(t_file_info))) {
        strcpy(file_info[file_num]->name, strupr(dirinfo.ff_name));
        file_info[file_num]->size = dirinfo.ff_fsize;
        file_info[file_num]->dir = (dirinfo.ff_attrib & FA_DIREC) == FA_DIREC;
        if (!strcmp(get_extension(file_info[file_num]->name),"SMD"))
          file_info[file_num]->rom = _SMD;
        else if(!strcmp(get_extension(file_info[file_num]->name),"BIN"))
          file_info[file_num]->rom = _BIN;
        else if(!strcmp(get_extension(file_info[file_num]->name),"1"))
          file_info[file_num]->rom = _SPLIT;
        else file_info[file_num]->rom = false;
        file_num++;
      }
    }
}

void free_file_info()
/* frees all memory used by the entries of the file_info structure */
{
  int i;
  for (i=0;i<file_num; i++) {
    if (file_info[i])
      free(file_info[i]);
  }
}

void get_files()
/* scan current directory for files */
/* the found file names and according infos are stored */
/* informations in array file_info */
/* All directories are found, while normal files are masked by */
/* the wildcard str_wild */
{
  char* wildcards[] = { "*.BIN", "*.1" };
  int i;

  free_file_info();
  file_num = 0;
  /* first get only directories */
  doserror=findfirst(wild_all,&dirinfo,FA_MYFILE +FA_DIREC);
  if (!doserror) {
    if (dirinfo.ff_attrib & FA_DIREC) add_file();
    do {
      doserror=findnext(&dirinfo);
      if (!doserror)
        if (dirinfo.ff_attrib & FA_DIREC) add_file();
    } while (!doserror && file_num<maxfiles);
  }
  /* now get rest of files */
  doserror=findfirst(str_wild,&dirinfo,FA_MYFILE);
  if (!doserror) {
    add_file();
    do {
      doserror=findnext(&dirinfo);
      if (!doserror) add_file();
    } while (!doserror && file_num<maxfiles);
  }
  /* now get also *.BIN and *.1 if searched for *.SMD files */
  if (wild==w_rom) {
    for (i=0; i<2; i++) {
      doserror=findfirst(wildcards[i],&dirinfo,FA_MYFILE);
      if (!doserror) {
        add_file();
        do {
          doserror=findnext(&dirinfo);
          if (!doserror) add_file();
        } while (!doserror && file_num<maxfiles);
      }
    }
  }
  if (file_num>1) sort_files();
}

void code_block(block16k _from, block16k _to)
/* converts 16kb noninterleaved data into interleaved data */
{
  int i;

  for (i=0; i<8192; i++) {
    _to[0x2000+i] = _from[2*i];
    _to[i]        = _from[2*i+1];
  }
}

void decode_block(block16k _from, block16k _to)
/* converts 16kb interleaved data into noninterleaved data */
{
  int i;
  for (i=0; i<8192; i++) {
    _to[2*i]   = _from[0x2000+i];
    _to[2*i+1] = _from[i];
  }
}

void clean_header()
/* initializes all displayed data to start values */
{
  str_overseas[0]=str_domestic[0]=str_copyright[0]=str_production[0]=
    str_control[0]=str_country[0]=str_checksum[0]=checksum=str_blocks[0]=
      japan=usa=europe=0;
}

int has_header(int idx)
/* decide if rom file has header by checking its size */
{
  int h;
  h = file_info[idx]->size % 16384;
  if (h==512) return true;                       /* yes, there is a header */
  else if (!h) return false;                  /* no, there ain't no header */
  else return error;                                      /* no rom at all */
}

void kill_space(char s[])
/* kills the trailing spaces of a string */
{
  int i;
  for (i=strlen(s)-1;i>=0;i--)
    if (s[i]==32) s[i]=0;
    else break;
}

void format_crc(char s_crc[], word i_crc)
/* converts crc from big endian to little endian (or vice versa ? :) */
/* then converts to string and formats to 4 uppercase characters */
{
  int i;
  word swap_crc;
  char str_help[10];

  swap_crc = swap_hilo(i_crc);
  itoa(swap_crc,str_help,16);
  s_crc[0]=0;
  for (i=0; i<4-strlen(str_help);i++)
    strcat(s_crc,"0");
  strcat(s_crc,str_help);
  strupr(s_crc);
}

int in_string(char c, char s[])
/* checks if char c is in string s */
{
  int k;
  int found=false;

  for (k=0; k<strlen(s); k++)
    if (c==s[k]) {
      found=true;
      break;
    }
  return found;
}

int get_header(char fn[])
/* read embedded header of file named fn */
/* automatically skips SMD-header and convert from */
/* interleaved to noninterleaved */
{
  byte h_blocks;
  block16k help_buf;
  dword romsize;
  char str_int[10];

  f = fopen(fn,"rb");
  if (f==NULL) {
    ok_alert(io_err,0,0);
    strcpy(str_info,io_err_inf);
    clean_header();
    return false;
  }
  if (smd || split) {
    h_blocks=fgetc(f);                     /* read block no. in SMD header */
    fseek(f,512,SEEK_SET);                              /* skip SMD header */
  }
  fread(&header_buf,1,16384,f);                   /* read first 16kb block */
  fclose(f);
  if (smd || split) {
    decode_block(header_buf,help_buf);
    memcpy(header_buf,help_buf,16384);
  }
  geheader =(t_header*)(&header_buf);               /* set header bufstart */
  geheader++;                                      /* set header at 0x100 */
  /* extract header info into strings */
  memcpy(str_overseas,geheader->overseas,48);     str_overseas[48]='\0';
  kill_space(str_overseas);
  memcpy(str_domestic,geheader->domestic,48);     str_domestic[48]='\0';
  kill_space(str_domestic);
  memcpy(str_copyright,geheader->copyright,16);   str_copyright[16]='\0';
  memcpy(str_production,geheader->production,14); str_production[14]='\0';
  checksum = swap_hilo(geheader->checksum);
  format_crc(str_checksum,checksum);
  memcpy(str_control,geheader->control,16); str_control[16]=0;
  memcpy(str_country,geheader->country,16); str_country[16]=0;
  kill_space(str_country);
  japan  = (in_string('J',str_country));
  usa    = (in_string('U',str_country));
  europe = (in_string('E',str_country));
  romsize = swap_dword(geheader->rom_end) - swap_dword(geheader->rom_start);
  romsize++;
  rblocks = romsize / 16384;
  itoa(blocks,str_blocks,10);
  strcat(str_blocks,"/");
  itoa(rblocks,str_int,10);
  strcat(str_blocks,str_int);
  if (smd || split)
    if (h_blocks != blocks)                    /* size in header correct ? */
      if (!(blocks==256 && !h_blocks)) {
        /* 16MBit roms are 256*16kb -> h_blocks = 0 !!! */
        strcat(str_blocks,"<smd header fixed>");
        changed=true;
      }
  return true;
}

int check_rom()
/* check if selected file named "str_fname" is rom */
{
  int header;
  int rom;

  changed = 0;
  strcpy(str_fname,file_info[file_idx]->name);
  rom = file_info[file_idx]->rom;
  if (rom) {
    blocks = file_info[file_idx]->size / 16384;
    header = has_header(file_idx);
    if ((header==true) && (rom==_SMD)) {
      bin=false; smd=true; split = false; 
    }
    else if (!header && (rom==_BIN)) {
      bin=true; smd=false; split = false; 
    }
    else if (rom==_SPLIT) {
      bin=false; smd=false; split=true;
    }
    else if (header == error) {
      strcpy(str_info,"<wrong size>");
      clean_header();
      return false;    /* size is not correct */
    }
  }
  else {
    strcpy(str_info,"<no rom>");
    clean_header();
    return false;
  }
  /* ok, it is a valid rom */
  if (get_header(str_fname)) {
    if (smd) strcpy(str_info,"Genesis Rom (SMD)");
    else if (bin) strcpy(str_info,"Genesis Rom (BIN)");
    else strcpy(str_info,"Genesis Rom (SPLIT)");
    return true;
  }
  else return false;
}

void check_disabled()
/* change/set state of displayed dialog parts */
{
  /* check header info */
  if (bin || smd || split) {
    my_dialog[text_overseas].flags    &= ~D_DISABLED;
    my_dialog[edit_overseas].flags    &= ~D_DISABLED;
    my_dialog[text_domestic].flags    &= ~D_DISABLED;
    my_dialog[edit_domestic].flags    &= ~D_DISABLED;
    my_dialog[text_blocks_].flags     &= ~D_DISABLED;
    my_dialog[text_blocks].flags      &= ~D_DISABLED;
    my_dialog[text_copyright_].flags  &= ~D_DISABLED;
    my_dialog[text_copyright].flags   &= ~D_DISABLED;
    my_dialog[text_production_].flags &= ~D_DISABLED;
    my_dialog[text_production].flags  &= ~D_DISABLED;
    my_dialog[text_checksum_].flags   &= ~D_DISABLED;
    my_dialog[text_checksum].flags    &= ~D_DISABLED;
    my_dialog[text_control_].flags    &= ~D_DISABLED;
    my_dialog[text_control].flags     &= ~D_DISABLED;
    my_dialog[text_country].flags     &= ~D_DISABLED;
    my_dialog[check_country_j].flags  &= ~D_DISABLED;
    my_dialog[check_country_u].flags  &= ~D_DISABLED;
    my_dialog[check_country_e].flags  &= ~D_DISABLED;
  }
  else {
    my_dialog[text_overseas].flags    |= D_DISABLED;
    my_dialog[edit_overseas].flags    |= D_DISABLED;
    my_dialog[text_domestic].flags    |= D_DISABLED;
    my_dialog[edit_domestic].flags    |= D_DISABLED;
    my_dialog[text_blocks_].flags     |= D_DISABLED;
    my_dialog[text_blocks].flags      |= D_DISABLED;
    my_dialog[text_copyright_].flags  |= D_DISABLED;
    my_dialog[text_copyright].flags   |= D_DISABLED;
    my_dialog[text_production_].flags |= D_DISABLED;
    my_dialog[text_production].flags  |= D_DISABLED;
    my_dialog[text_checksum_].flags   |= D_DISABLED;
    my_dialog[text_checksum].flags    |= D_DISABLED;
    my_dialog[text_control_].flags    |= D_DISABLED;
    my_dialog[text_control].flags     |= D_DISABLED;
    my_dialog[text_country].flags     |= D_DISABLED;
    my_dialog[check_country_j].flags  |= D_DISABLED;
    my_dialog[check_country_u].flags  |= D_DISABLED;
    my_dialog[check_country_e].flags  |= D_DISABLED;
  }
  /* check rom functions */
  if (bin || smd) {
    my_dialog[button_run].flags       &= ~D_DISABLED;
    my_dialog[button_convert].flags   &= ~D_DISABLED;
    my_dialog[button_calc_crc].flags  &= ~D_DISABLED;
    my_dialog[button_lsetup].flags    &= ~D_DISABLED;
  }
  else {
    my_dialog[button_run].flags       |= D_DISABLED;
    my_dialog[button_convert].flags   |= D_DISABLED;
    my_dialog[button_calc_crc].flags  |= D_DISABLED;
    my_dialog[button_lsetup].flags    |= D_DISABLED;
  }
  /* check write button */
  if (changed) my_dialog[button_write].flags &= ~D_DISABLED;
  else my_dialog[button_write].flags |= D_DISABLED;
  /* country checks */
  if (japan) my_dialog[check_country_j].flags |= D_SELECTED;
  else my_dialog[check_country_j].flags &= ~D_SELECTED;
  if (usa) my_dialog[check_country_u].flags |= D_SELECTED;
  else my_dialog[check_country_u].flags &= ~D_SELECTED;
  if (europe) my_dialog[check_country_e].flags |= D_SELECTED;
  else my_dialog[check_country_e].flags &= ~D_SELECTED;
  /* radio buttons */
  if (wild==w_rom) {
    my_dialog[radio_rom].flags    |= D_SELECTED;
    my_dialog[radio_all].flags    &= ~D_SELECTED;
  }
  else if (wild==w_all) {
    my_dialog[radio_rom].flags    &= ~D_SELECTED;
    my_dialog[radio_all].flags    |= D_SELECTED;
  }
  /* rename/delete buttons */
  if (str_fname[0]) {
    my_dialog[button_delete].flags &= ~D_DISABLED;
    my_dialog[button_rename].flags &= ~D_DISABLED;
  }
  else {
    my_dialog[button_delete].flags |= D_DISABLED;
    my_dialog[button_rename].flags |= D_DISABLED;
  }
  /* join button */
  if (split)
    my_dialog[button_join].flags  &= ~D_DISABLED;
  else
    my_dialog[button_join].flags  |= D_DISABLED;
}

void fill_space(char s[], int size)
/* fill end of string with spaces until it reaches size */
{
  int i;
  int l=strlen(s);

  for (i=l;i<size;i++) {
    s[i]=32;
  }
}

int write_header()
/* writes embedded header to currently selected file named str_fname */
/* automatically skips SMD-header and converts interleaving */
{
  block16k help_buf;
  byte h_blocks;
  dword romsize;

  /* copy all info into header */
  memcpy(geheader->overseas,str_overseas,48);
  fill_space(geheader->overseas,48);
  memcpy(geheader->domestic,str_domestic,48);
  fill_space(geheader->domestic,48);
  memcpy(geheader->copyright,str_copyright,16);
  fill_space(geheader->copyright,16);
  memcpy(geheader->production,str_production,14);
  fill_space(geheader->production,14);
  geheader->checksum = swap_hilo(checksum);
  memcpy(geheader->control,str_control,16);
  fill_space(geheader->control,16);
  str_country[0]=0;
  if (japan)  strcat(str_country,"J");
  if (usa)    strcat(str_country,"U");
  if (europe) strcat(str_country,"E");
  memcpy(geheader->country,str_country,16);
  fill_space(geheader->country,16);
  romsize = rblocks * 16384;
  romsize--;
  geheader->rom_start = 0;
  geheader->rom_end = swap_dword(romsize);
  /* open file */
  f = fopen(str_fname,"rb+");
  if (f==NULL) {
    ok_alert(io_err,0,0);
    strcpy(str_info,io_err_inf);
    return false;
  }
  /* write data */
  if (smd) {
    h_blocks=blocks & 0xff;
    fputc(h_blocks,f);                           /* write correct block.no */
    fseek(f,512,SEEK_SET);                              /* skip SMD header */
    code_block(header_buf,help_buf);          /* code block to interleaved */
    fwrite(&help_buf,1,16384,f);             /* smd: write first 16k block */
  }
  else fwrite(&header_buf,1,16384,f);        /* bin: write first 16k block */
  fclose(f);
  changed = false;
}

void conv_slash(char p[])
/* converts slashs in string into backslashs */
{
  int i;

  for (i=0; i<strlen(p); i++)
    if (p[i]=='/') p[i]='\\';
}

int check_path(char *path, int pop_alert)
/* checks if the given path is valid */
/* if pop_alert is true, an alert will pop up if the path is invalid */
{
  char testpath[128];

  strcpy(testpath,strupr(path));
  put_backslash(testpath);
  strcat(testpath,"*.*");
  if (!file_exists(testpath, 63, NULL)) {
    if (pop_alert)
      ok_alert("The path or drive",path,"does not exist.");
    getcwd(path,127);
    conv_slash(path);
    return(false);
  }
  return(true);
}

void change_path(char *path)
/* changes the path after checking it */
  {
  if (check_path(path, TRUE)) {
    chdir(path);
    getcwd(path,127);
    conv_slash(path);
    get_files();
  }
}

void clip_path(char p[])
/* deletes filename from a path */
/* e.g. "c:\emu\megatool.exe" is converted to "c:\emu" */
/* but "c:\megatool.exe" is converted to "c:\" !!! */
{
  char s[128];
  int sl;
  int i;

  strcpy(s,p);
  sl = strlen(s);
  for (i=sl-1; i>=0; i--) {
    if (s[i]==':')  break;
    if (s[i]=='\\') break;
    if (s[i]=='/') break;
  }
  if (i>0)
   if(s[i-1]==':')
     i++;
  p[i]=0;
}

kill_extension(char s[])
/* kill the last point in a name and thereby the extension */
{
  int i;
  int j;

  j=strlen(s)-1;
  for (i=j; i >= 0; i--)
    if (s[i]=='.') {
      s[i]=0;
      break;
    }
}

int file_request(char title[], char name[], char wild[])
/* show a file request box */
/* title is shown at top of the box */
/* name is path+filename to start with - path must be specified !*/
/* wild is wildcard to start with */
{
  int fr_leave=false;
  int rvalue=false;
  char save_wild[13];

  BITMAP *bmp;

  bmp = create_bitmap(fr_dialog->w+1, fr_dialog->h+1);

  /* save background */
  if (bmp) {
    show_mouse(NULL);
    blit(screen,bmp,fr_dialog->x,fr_dialog->y,0,0,fr_dialog->w+1,fr_dialog->h+1);
  }

  strcpy(fr_fname,get_filename(name));
  strcpy(fr_path,name);
  if (fr_fname[0]) clip_path(fr_path);
  strcpy(fr_title,title);
  strcpy(save_wild,str_wild);
  strcpy(str_wild,wild);
  change_path(fr_path);
  while(!fr_leave) {
    ret = do_dialog( fr_dialog, find_dialog_focus(fr_dialog));
    switch( ret ) {
      case fr_button_ok:
        fr_leave=true;
        strcpy(name,fr_path);
        put_backslash(name);
        strcat(name,fr_fname);
        rvalue=true;
        break;
      case fr_button_cncl:
        fr_leave=true; break;
        break;
      case fr_button_a:
        strcpy(fr_path,"A:");
        change_path(fr_path);
        break;
      case fr_button_b:
        strcpy(fr_path,"B:");
        change_path(fr_path);
        break;
      case fr_button_c:
        strcpy(fr_path,"C:");
        change_path(fr_path);
        break;
      case fr_button_d:
        strcpy(fr_path,"D:");
        change_path(fr_path);
        break;
      case fr_button_e:
        strcpy(fr_path,"E:");
        change_path(fr_path);
        break;
      case fr_button_f:
        strcpy(fr_path,"F:");
        change_path(fr_path);
        break;
      case fr_button_g:
        strcpy(fr_path,"G:");
        change_path(fr_path);
        break;
      case fr_button_h:
        strcpy(fr_path,"H:");
        change_path(fr_path);
        break;
      case fr_edit_path:
        change_path(fr_path);
        break;
      case fr_edit_name:
        if (fr_fname[0]==0) strcpy(str_wild,wild_all);
        else strcpy(str_wild,fr_fname);
        get_files();
        break;
      case fr_list:
        file_idx = fr_dialog[fr_list].d1;
        if (file_info[file_idx]->dir) {
          chdir(file_info[file_idx]->name);
          getcwd(fr_path,127);
          conv_slash(fr_path);
          get_files();
          my_dialog[file_list].d1 = 0; my_dialog[file_list].d2 = 0;
        }
        else {
          strcpy(fr_fname,file_info[file_idx]->name);
        }
        break;
    }
  }
  /* restore background */
  if (bmp) {
    show_mouse(NULL);
    blit(bmp,screen,0,0,fr_dialog->x, fr_dialog->y, fr_dialog->w+1, fr_dialog->h+1);
    destroy_bitmap(bmp);
  }
  strcpy(str_wild,save_wild);
  change_path(str_path);
  return rvalue;
}

void sort_list()
/* sort lines in text_list */
{
  char* hstring;
  int i=1;
  int j=1;

  do {
    for (i= j+1; i< line_num;i++)
      if (strcmp(text_list[j],text_list[i])>0) {
        hstring = text_list[j];
        text_list[j]=text_list[i];
        text_list[i]=hstring;
      }
    j++;
  } while (j < line_num);
}

void free_text_list()
/* frees all memory occupied by entries of text_list */
{
  int i;
  for (i=0;i<line_num; i++) {
    if (text_list[i])
      free(text_list[i]);
  }
}

int create_list()
/* create rom list */
{
  FILE *fl;
  char s_out[128];
  char s_help[128];
  char s_fname[128];
  int j;
  int h;
  int header;

  line_num=0;
  show_mouse(NULL);
  set_mouse_sprite(&spr_wait);
  show_mouse(screen);
  for (i=0; i<file_num; i++)
    if (file_info[i]->rom) {
      header=has_header(i);
      if ((header==true) && (file_info[i]->rom==_SMD)) {
        bin=false; smd=true;
      }
      else if (!header && (file_info[i]->rom==_BIN)) {
        bin=true; smd=false;
      }
      else {
        bin =false; smd = false;
      }
      if (bin||smd)
        if (get_header(file_info[i]->name)) {
          strcpy(s_out,str_overseas);
          h=65-strlen(s_out);
          for (j=0;j<h;j++)
            strcat(s_out," ");
          strcpy(s_help,str_country);
          h=3-strlen(str_country);
          if (h>0)
            for (j=0;j<h;j++)
              strcat(s_help," ");
          s_help[3]=0;
          strcat(s_out, s_help);
          strcat(s_out,"      ");
          itoa(file_info[i]->size/1024,s_help,10);
          for (j=0; j<4-strlen(s_help); j++) strcat(s_out," ");
          strcat(s_out,s_help);
          if ((text_list[line_num]=(char*)malloc(90))&&(line_num<maxlines)) {
            strcpy(text_list[line_num],s_out);
            line_num++;
          }
        }
    }
  sort_list();
  /* open list file */
  show_mouse(NULL);
  set_mouse_sprite(&spr_mouse);
  show_mouse(screen);
  check_path(str_path,FALSE);
  strcpy(s_fname,str_path);
  put_backslash(s_fname);
  strcat(s_fname,"megatool.lst");
  if (!file_request("GenEm rom list",s_fname,"*.*")) return false;
  if (!file_overwrite(s_fname))
    return false;
  fl = fopen(s_fname,"wt");
  if (fl==NULL) {
    ok_alert("Couldn't open","list file !",0);
    return false;
  }
  /* write list file */
  fputs("Name                                                             Country  KB\n",fl);
  fputs("-------------------------------------------------------------------------------\n",fl);
  for (i=0; i<line_num; i++) {
    fputs(text_list[i],fl);
    fputs("\n",fl);
  }
  fclose(fl);
  free_text_list();
}

void init_allegro( void )
/*  Get Allegro started. */
{
  allegro_init();
  debug("allegro initialized");
  install_keyboard_hooks(replacement_keypressed,replacement_readkey);
  if (install_mouse()!=false)                      /*  You need the mouse. */
    debug("mouse installed");
  else
    debug("no mouse found");
  set_mouse_sprite(&spr_mouse);
  install_timer();                     /*  And the mouse needs the timer.  */
  debug("timer installed");
  if (set_gfx_mode( GFX_VESA2L, 640, 480, 0, 0 ) <0)
    if (set_gfx_mode( GFX_VESA2B, 640, 480, 0, 0 ) <0)
      if (set_gfx_mode( GFX_VESA1, 640, 480, 0, 0 )<0)
        if (set_gfx_mode( GFX_AUTODETECT, 640, 480, 0, 0 )<0) {
          debug(allegro_error);
          exit(1);
        }
        else debug("autodetect gfx mode set");
      else debug("vesa1.2 gfx mode set");
    else debug("vesa2.0b gfx mode set");
  else debug("vesa2.0l gfx mode set");
  set_palette(pal);
  gui_fg_color = 1;
  gui_mg_color = 24;
  gui_bg_color = 7;
  clr_scrn();
}

void run_game()
/* starts GenEm with selected rom */
{
  char exec_path[128];
  char str_int[5];
  word start_option;

  int option = 0;
  /* alert if header was changed */
  if (changed) {
    if (alert3d("Write Header","or discard changes",0,"&Write","&Discard",0,'w','d',0)==1)
      write_header();
    else {
      bin=smd=blocks=rblocks=changed=0;
      check_rom();
    }
  }
  /* check if executable exists */
  if (!file_exists(str_genem, 63, NULL))
    ok_alert("GenEm executable","not found in the given path:",str_genem);
  else {
    /* put together full rom name */
    strcpy(str_exec,str_path);
    put_backslash(str_exec);
    strcat(str_exec,str_fname);
    /* change to the executables directory */
    strcpy(exec_path,str_genem);
    clip_path(exec_path);
    chdir(exec_path);
    /* close down allegro */
    allegro_exit();
    /* check if local option exists, else use global options */
    strcpy(str_name,str_fname);
    if (strlen(str_name)>3)
      kill_extension(str_name);
    opt_idx = index_local_option();
    if (opt_idx==-1) start_option = g_option;                    /* global */
    else start_option = local_option[opt_idx]->option;            /* local */
    /* put together options */
    options[option]=str_genem;             /* option zero is path to GenEm */
    option++;
    itoa((start_option&7)+1,str_int,10);             /* option: frame skip */
    options[option]="-frame";
    option++;
    options[option]=str_int;
    option++;
    if (!usa) {                          /* set country specific behaviour */
      if (europe) {
        options[option]="-europe";
        option++;
      }
      else if (japan) {
        options[option]="-japan";
        option++;
      }
    }
    if (start_option&_modex) {                           /* option: mode x */
      options[option]="-modex";
      option++;
    }
    if (start_option&_nojoy) {                            /* option: nojoy */
      options[option]="-nojoy";
      option++;
    }
    if (start_option&_nosound) {                        /* option: nosound */
      options[option]="-nosound";
      option++;
    }
    if (start_option&_z80) {                                /* option: z80 */
      options[option]="-z80";
      option++;
    }
    if (start_option&_fakedz80) {                     /* option: faked z80 */
      options[option]="-fakedz80";
      option++;
    }
    options[option]=str_exec;
    option++;
    options[option]=0;
    /* execute */
    set_gmode();
    __djgpp_exception_toggle();
    spawnv(P_WAIT,str_genem,options);
    __djgpp_exception_toggle();
    /* get back to old path and start allegro again */
    init_allegro();
    change_path(str_path);
    check_rom();
  }
}

void option_string(word opt, char s[])
/* convert option bits into readable string format used in ini files */
{
  s[0] = (opt&7)+49;        /* +48 to convert to ASCII, +1 to start at '1' */
  if (opt&_modex)    s[1] = 'y'; else s[1]='n';
  if (opt&_nojoy)    s[2] = 'y'; else s[2]='n';
  if (opt&_nosound)  s[3] = 'y'; else s[3]='n';
  if (opt&_z80)      s[4] = 'y'; else s[4]='n';
  if (opt&_fakedz80) s[5] = 'y'; else s[5]='n';
  s[6]=0;
}

int write_cfg()
/* write global settings to file */
{
  char str_help[128];

  strcpy(str_help,str_ini);
  put_backslash(str_help);
  strcat(str_help,"MegaTool.cfg");
  f = fopen(str_help,"wt");
  if (f==NULL) {
    ok_alert(io_err,"Couldn't write","Cfg File !!!");
    return false;
  }
  fputs(ident_exec,f); fputs("\n",f);
  fputs(str_genem,f); fputs("\n",f);
  fputs(ident_startpath,f); fputs("\n",f);
  fputs(str_startpath,f); fputs("\n",f);
  fputs(ident_romsonly,f); fputs("\n",f);
  itoa(roms_only,str_help,10);
  fputs(str_help,f); fputs("\n",f);
  fputs(ident_g_options,f); fputs("\n",f);
  option_string(g_option,str_help);
  fputs(str_help,f); fputs("\n",f);
  fclose(f);
}

char *_fgets(char *buffer, int maxlength, FILE *file)
/* modified fgets that doesn't return the \n at the line's end */
{
  char* _ret;
  _ret = fgets(buffer, maxlength, file);
  buffer[strlen(buffer)-1]=0;
  return _ret;
}

word parse_options(char s[])
/* stores options from option string s in a 16bit internal */
/* format and gives back that value */
{
  char str_help[21];
  word opt=0;
  int i;
  int h;

  /* make sure the string has a defined size */
  strncpy(str_help,s,20);
  h = strlen(s);
  if  (h < 20)
    for (i=h;i<20;i++)
      str_help[i]=32;
  /* parse string */
  opt = str_help[0]-49;                  /* get frame skip and store it in */
  if (opt>7) opt=7;                                  /* the lowest 3 bits  */
  if (str_help[1]=='y') opt|= _modex;
  if (str_help[2]=='y') opt|= _nojoy;
  if (str_help[3]=='y') opt|= _nosound;
  if (str_help[4]=='y') opt|= _z80;
  if (str_help[5]=='y') opt|= _fakedz80;
  return opt;
}

int read_cfg()
/* read global settings from file */
{
  char str_help[128];

  strcpy(str_help,str_ini);
  put_backslash(str_help);
  strcat(str_help,"MegaTool.cfg");
  f = fopen(str_help,"rt");
  if (f!=NULL) {
    while (_fgets(str_help,127,f)!=NULL) {
      if (!strncmp(str_help,ident_exec,strlen(ident_exec)))
        _fgets(str_genem,127,f);
      if (!strncmp(str_help,ident_startpath,strlen(ident_startpath)))
        _fgets(str_startpath,127,f);
      if (!strncmp(str_help,ident_romsonly,strlen(ident_romsonly))) {
        _fgets(str_help,127,f);
        roms_only = atoi(str_help);
      }
      if (!strncmp(str_help,ident_g_options,strlen(ident_g_options))) {
        _fgets(str_help,20,f);
        g_option = parse_options(str_help);
      }
    }
    fclose(f);
  }
  if (roms_only) {
    wild= w_rom;
    strcpy(str_wild,"*.SMD");
  }
  else {
    wild= w_all;
    strcpy(str_wild,wild_all);
  }
}

int check_setup()
/* checks the values given in str_startpath and str_genem */
/* used by setup() */
{
  char str_help[128];

  if (!file_exists(str_genem, 63, NULL))
    if (alert3d("The entered executable doesn't exist !",
         "Do you wish to set it anyway ???",0,"&Yes","&No",0,'y','n',0)==2)
      return false;
  strcpy(str_help,str_startpath);   /* use different var since check_path */
  if (!check_path(str_help,false))     /* corrects the path automatically */
    if (alert3d("The entered Startpath doesn't exist!",
         "Do you wish to set it anyway ???",0,"&Yes","&No",0,'y','n',0)==2)
      return false;
  return true;
}

get_frame(word *f, char s[])
/* converts a character number '1'-'8' from s into a binary number 0-7 */
/* and puts it into the least significant bits of f */
{
  char c;

  c= s[0];
  if (c<'1') c='1';
  if (c>'8') c='8';
  s[0]=c;
  *f &= ~7;
  *f |= c-49;
}

void global_setup()
/* implements the global setup dialog */
{
  int s_leave=false;
  char c;
  BITMAP *bmp;

  bmp = create_bitmap(setup_dialog->w+1, setup_dialog->h+1);

  /* save background */
  if (bmp) {
    show_mouse(NULL);
    blit(screen,bmp,setup_dialog->x,setup_dialog->y,0,0,setup_dialog->w+1,
      setup_dialog->h+1);
  }

  while(!s_leave) {
    /* init check boxes */
    str_frame[0]=(g_option&7)+49; /* +48 convert to ASCII, +1 start at '1' */
    str_frame[1]=0;

    if (roms_only) setup_dialog[setup_check_romsonly].flags  |= D_SELECTED;
    else setup_dialog[setup_check_romsonly].flags  &= ~D_SELECTED;

    if (g_option&_modex)
      setup_dialog[setup_check_modex].flags  |= D_SELECTED;
    else setup_dialog[setup_check_modex].flags  &= ~D_SELECTED;
    if (g_option&_nojoy)
      setup_dialog[setup_check_nojoy].flags  |= D_SELECTED;
    else setup_dialog[setup_check_nojoy].flags  &= ~D_SELECTED;
    if (g_option&_nosound)
      setup_dialog[setup_check_nosound].flags  |= D_SELECTED;
    else setup_dialog[setup_check_nosound].flags  &= ~D_SELECTED;
    if (g_option&_z80)
      setup_dialog[setup_check_z80].flags  |= D_SELECTED;
    else setup_dialog[setup_check_z80].flags  &= ~D_SELECTED;
    if (g_option&_fakedz80)
      setup_dialog[setup_check_fakedz80].flags  |= D_SELECTED;
    else setup_dialog[setup_check_fakedz80].flags  &= ~D_SELECTED;

    ret = do_dialog( setup_dialog, find_dialog_focus(setup_dialog));
    get_frame(&g_option,str_frame);
    switch( ret ) {
      case setup_button_ok:
        if (check_setup()) s_leave=true;
        break;
      case setup_button_save:
        write_cfg();
        if (check_setup()) s_leave=true;
        break;
      case setup_check_romsonly:
        roms_only=!roms_only;
        break;
      case setup_check_modex:
        g_option^=_modex;
        break;
      case setup_check_nojoy:
        g_option^=_nojoy;
        break;
      case setup_check_nosound:
        g_option^=_nosound;
        break;
      case setup_check_z80:
        g_option^=_z80;
        break;
      case setup_check_fakedz80:
        g_option^=_fakedz80;
        break;
    }
  }
  /* restore background */
  if (bmp) {
    show_mouse(NULL);
    blit(bmp,screen,0,0,setup_dialog->x, setup_dialog->y, setup_dialog->w+1,
      setup_dialog->h+1);
    destroy_bitmap(bmp);
  }
}

int add_local_option()
/* add a local option entry to the list */
{
  if (opt_num < maxoptions)
    if (local_option[opt_num] = (t_local_option*)malloc(sizeof(t_local_option))) {
      strncpy(local_option[opt_num]->name, str_name,8);
      local_option[opt_num]->name[8]=0;
      local_option[opt_num]->option = 0;
      opt_num++;
      return true;
    }
    else return false;
  else return false;
}

void kill_local_option(int index)
/* kill a local option entry from the list */
{
  int i;

  if (local_option[index]) {
    free(local_option[index]);
    for (i=index; i<opt_num-1; i++)
      local_option[i] = local_option[i+1];
    if (opt_num>0) opt_num--;
    local_option[opt_num]=0;
  }
}

int index_local_option()
/* find a local option entry in the list and return its index */
{
  int i;
  int found=-1;

  for (i=0; i<opt_num; i++)
    if (!strcmp(str_name,local_option[i]->name)) found = i;
  return found;
}

int write_local_setup()
/* write local settings to file */
{
  char str_help[128];
  int i;

  strcpy(str_help,str_ini);
  put_backslash(str_help);
  strcat(str_help,"MegaTool.ini");
  f = fopen(str_help,"wt");
  if (f==NULL) {
    ok_alert(io_err,"Couldn't write","Ini File !!!");
    return false;
  }
  fputs("***** format: name@frame|modex|nojoy|nosound|z80|fakedz80 *****",f);
  fputs("\n",f);
  for (i=0; i<opt_num; i++) {
    fputs(local_option[i]->name,f);
    fputs("@",f);
    option_string(local_option[i]->option,str_help);
    fputs(str_help,f);
    fputs("\n",f);
  }
  fclose(f);
}

void read_local_setup()
/* read local settings from file */
{
  char str_help[128];
  int index=-1;

  strcpy(str_help,str_ini);
  put_backslash(str_help);
  strcat(str_help,"MegaTool.ini");
  f = fopen(str_help,"rt");
  if (f!=NULL) {
    while (_fgets(str_help,127,f)!=NULL) {
      if (str_help[0]!='*') {
        for (i=0; i< strlen(str_help); i++)
          if (str_help[i]=='@') index=i;
        strncpy(str_name,str_help,index);
        str_name[index]=0;
        strupr(str_name);
        if (add_local_option())
          local_option[opt_num-1]->option=
            parse_options((char*)&str_help[index+1]);
      }
    }
    fclose(f);
  }
}

int local_setup()
/* implements the local setup dialog */
{
  int s_leave=false;
  char c;
  word opt;
  BITMAP *bmp;

  /* build name */
  strcpy(str_name,str_fname);
  strupr(str_name);
  if (strlen(str_name)>3)
    kill_extension(str_name);
  /* look if local option for this name exists */
  if ((opt_idx=index_local_option())==-1)
    if (alert3d("No local option exists for",str_name,"Create One ?",
      "&Yes","&No",0,'y','n',0)==1)
      if (add_local_option()) opt_idx=opt_num-1;
  if (opt_idx==-1) return false;
  /* ok, let's start */
  bmp = create_bitmap(setup_dialog->w+1, setup_dialog->h+1);
  /* save background */
  if (bmp) {
    show_mouse(NULL);
    blit(screen,bmp,setup_dialog->x,setup_dialog->y,0,0,setup_dialog->w+1,
      setup_dialog->h+1);
  }

  opt = local_option[opt_idx]->option;
  while(!s_leave) {
    /* init check boxes */
    str_frame[0]=(opt&7)+49; /* +48 convert to ASCII, +1 start at '1' */
    str_frame[1]=0;

    if (opt&_modex)
      lsetup_dialog[lsetup_check_modex].flags  |= D_SELECTED;
    else lsetup_dialog[lsetup_check_modex].flags  &= ~D_SELECTED;
    if (opt&_nojoy)
      lsetup_dialog[lsetup_check_nojoy].flags  |= D_SELECTED;
    else lsetup_dialog[lsetup_check_nojoy].flags  &= ~D_SELECTED;
    if (opt&_nosound)
      lsetup_dialog[lsetup_check_nosound].flags  |= D_SELECTED;
    else lsetup_dialog[lsetup_check_nosound].flags  &= ~D_SELECTED;
    if (opt&_z80)
      lsetup_dialog[lsetup_check_z80].flags  |= D_SELECTED;
    else lsetup_dialog[lsetup_check_z80].flags  &= ~D_SELECTED;
    if (opt&_fakedz80)
      lsetup_dialog[lsetup_check_fakedz80].flags  |= D_SELECTED;
    else lsetup_dialog[lsetup_check_fakedz80].flags  &= ~D_SELECTED;

    ret = do_dialog( lsetup_dialog, find_dialog_focus(lsetup_dialog));
    get_frame(&opt,str_frame);
    switch( ret ) {
      case lsetup_button_remove:
        kill_local_option(opt_idx);
        s_leave=true;
        break;
      case lsetup_button_save:
        local_option[opt_idx]->option = opt;
        write_local_setup();
        s_leave=true;
        break;
      case lsetup_check_modex:
        opt^=_modex;
        break;
      case lsetup_check_nojoy:
        opt^=_nojoy;
        break;
      case lsetup_check_nosound:
        opt^=_nosound;
        break;
      case lsetup_check_z80:
        opt^=_z80;
        break;
      case lsetup_check_fakedz80:
        opt^=_fakedz80;
        break;
    }
  }
  /* restore background */
  if (bmp) {
    show_mouse(NULL);
    blit(bmp,screen,0,0,setup_dialog->x, setup_dialog->y, setup_dialog->w+1,
      setup_dialog->h+1);
    destroy_bitmap(bmp);
  }
}

int convert()
/* converts BIN<->SMD */
{
  char s_fname[128];                                        /* target name */
  char smd_header[512];                                 /* SMD file header */
  block16k from_buf;                      /* 16k block used for converting */
  block16k to_buf;
  FILE *ft;
  int i;

  /* open the source file */
  f = fopen(str_fname,"rb");
  if (f==NULL) {
    ok_alert(io_err,0,0);
    return false;
  }
  /* build target file name */
  strcpy(s_fname,str_path);
  put_backslash(s_fname);
  strcat(s_fname,str_fname);
  kill_extension(s_fname);
  if (bin) strcat(s_fname,".SMD");
  else strcat(s_fname,".BIN");
  /* let user edit target name */
  if (!file_request("Convert rom file",s_fname,"*.*")) {
    fclose (f);
    return false;
  }
  /* check if target file exists */
  if (!file_overwrite(s_fname)) {
    fclose (f);
    return false;
  }
  /* let's open target file */
  ft = fopen(s_fname,"wb");
  if (f==NULL) {
    ok_alert(io_err,0,0);
    fclose(f);
    return false;
  }
  show_mouse(NULL);
  set_mouse_sprite(&spr_wait);
  show_mouse(screen);
  if (bin) {
    /* init smd_header */
    for (i=0; i<512; i++)
      smd_header[i] = 0;
    smd_header[0]  = blocks;   /* no of 16kb blocks */
    smd_header[1]  = 3;        /* whatever this means */
    smd_header[8]  = 0xaa;     /* file ident */
    smd_header[9]  = 0xbb;     /* file ident */
    smd_header[10] = 6;        /* SMD ident */
    /* write smd_header */
    fwrite(smd_header,1,512,ft);
  }
  else
    fseek(f,512,SEEK_SET);                              /* skip SMD header */
  /* main loop */
  for (i=0; i<blocks; i++) {
    fread(from_buf,1,16384,f);
    if (smd)  decode_block(from_buf,to_buf);
    else code_block(from_buf,to_buf);
    fwrite(to_buf,1,16384,ft);
  }
  fclose(ft);
  fclose(f);
  show_mouse(NULL);
  set_mouse_sprite(&spr_mouse);
  show_mouse(screen);
}

void crc_block(block16k block, int ofs_s, dword *crc)
/* calculate the checksum of a 16kb block (added to existing crc) */
/* the only interesting thing here is the conversion from little */
/* endian to big endian (or the other way round..I just can't remember ;) */
{
  int i;

  for (i=ofs_s;i<16384;i+=2) {
    *crc+= block[i]<<8;
    *crc+= block[i+1];
  }
}

int calc_crc()
/* calculate checksum of selected file named str_fname */
{
  block16k read_buf;                      /* 16k block used for converting */
  block16k crc_buf;
  char str_help[15];
  int i;
  int ofs_start;
  dword new_crc=0;

  /* open the source file */
  f = fopen(str_fname,"rb");
  if (f==NULL) {
    ok_alert(io_err,0,0);
    return false;
  }
  if (smd)
    fseek(f,512,SEEK_SET);                              /* skip SMD header */
  show_mouse(NULL);
  set_mouse_sprite(&spr_wait);
  show_mouse(screen);
  /* main loop */
  for (i=0; i<blocks; i++) {
    if (!i) ofs_start=0x200;              /* first block: skip 0x200 bytes */
    else ofs_start=0;
    /* if (i==blocks-1) ofs_end = 16384; */
    fread(read_buf,1,16384,f);
    if (smd) {
      decode_block(read_buf,crc_buf);
      crc_block(crc_buf,ofs_start,&new_crc);
    }
    else  crc_block(read_buf,ofs_start,&new_crc);
  }
  fclose(f);
  show_mouse(NULL);
  set_mouse_sprite(&spr_mouse);
  show_mouse(screen);
  new_crc &= 0xffff;
  format_crc(str_help,new_crc);
  if (new_crc==checksum) ok_alert("Checksum",str_help,"is correct");
  else {
    ok_alert("Rom's checksum is wrong ! Correct checksum is",str_help,
      "Press 'Write Header' to fix rom.");
    checksum = new_crc;
    strcpy(str_checksum,str_help);
    strcat(str_checksum," <fixed>");
    changed=true;
  }
}

int _file_size(char name[])
/* returns false if no file of that name exists. */
/* else returns size of found file */
{
  struct ffblk dirinfo;

  if(!findfirst(name,&dirinfo,FA_MYFILE)) {
    return dirinfo.ff_fsize;
  }
  else return false;
}

int join_rom()
{
  FILE *fs;                                                 /* source file */
  FILE *ft;                                                 /* target file */
  int size;                                                   /* file size */
  int header;                                     /* boolean: has header ? */
  char ext[2];                          /* used for building the extension */
  char fn_source[255];                                 /* filename source  */
  char fn_target[255];                                  /* filename target */
  char str_help[13];                      /* used for building source name */
  char *buffer;                                   /* used for copying data */
  char smd_header[512];                                 /* SMD file header */

  /* create source name with full path */
  strcpy(fn_source,str_path);
  put_backslash(fn_source);
  strcat(fn_source,str_fname);
  /* create name without extension in str_help from source name */
  strcpy(str_help,str_fname);
  kill_extension(str_help);
  /* build default target name and get name with file requester */
  strcpy(fn_target,fn_source);
  kill_extension(fn_target);
  strcat(fn_target,".SMD");
  if (!file_request("Joined rom",fn_target,"*.*")) return false;
  strcat(str_help,".");                   /* fn_source contains "<source>." */
  /* check if target file exists */
  if (!file_overwrite(fn_target))
    return false;
  /* open target */
  ft= fopen(fn_target,"wb");
  if (ft==NULL) {
    ok_alert("Error while opening target file",0,0);
    return false;
  }
  /* init extension */
  ext[0] = '1';
  ext[1]= 0;
  blocks = 0;
  /* main loop */
  show_mouse(NULL);
  set_mouse_sprite(&spr_wait);
  show_mouse(screen);
  do {
    strcpy(fn_source,str_help);                 /* create source file name */
    strcat(fn_source,ext);
    /* check source file */
    size = _file_size(fn_source);
    if (size) {
      header = ((size % 1024) ==512);    /* does this file have a header ? */
      if (size%1024!=0 && !header) {
        fclose(ft);
        ok_alert("Invalid source file size:",fn_source,0);
        return false;
      }
      blocks += size / 16384;
      /* open source file */
      fs=fopen(fn_source,"rb");
      if (fs==NULL) {
        ok_alert("Error while opening source file...",0,0);
        fclose(ft);
        return false;
      }
      /* skip header if this in not the first file of the split archive */
      if (ext[0]!='1' && header) {
        fseek(fs,512,SEEK_SET);                             /* skip header */
        size -= 512;
      }
      /* allocate buffer */
      buffer = (char*)malloc(size);
      if (!buffer) {
        ok_alert(mem_err,0,0);
        fclose(ft);
        size = 0;
      }
      /* copy block */
      else {
        fread(buffer,1,size,fs);
        fwrite(buffer,1,size,ft);
        fclose(fs);
        free(buffer);
      }
    }
    ext[0]++;
  } while(size);
  show_mouse(NULL);
  set_mouse_sprite(&spr_mouse);
  show_mouse(screen);
  /* close target */
  if (has_header(file_idx)) {
    for (i=0; i<512; i++)
      smd_header[i] = 0;
    smd_header[0]  = blocks;   /* no of 16kb blocks */
    smd_header[1]  = 3;        /* whatever this means */
    smd_header[8]  = 0xaa;     /* file ident */
    smd_header[9]  = 0xbb;     /* file ident */
    smd_header[10] = 6;        /* SMD ident */
    /* write smd_header */
    fseek(ft,0,SEEK_SET);
    fwrite(smd_header,1,512,ft);
  }
  fclose(ft);
}

#define double_dump  1                                      /* double dump */
#define intro_dump   2                          /* dump has appended intro */
#define empty_dump 3                   /* dump has empty blocks at the end */
#define long_dump    4                                 /* dump is too long */
#define short_dump   5                                /* dump is too short */

int fix_rom(char str_1[], char str_2[], char str_3[], int diagnosis)
/* fix a rom with surplus pages by either truncating it or fixing */
/* the romsize info in the embedded header */
{
  int ret;
  byte b;
  dword fsize;

  if (diagnosis==short_dump)
    ret=alert3d(str_1,str_2,str_3,"&Cancel","Fix &Header",0,'c','h',0);
  else
    ret=alert3d(str_1,str_2,str_3,"&Cancel","Fix &Header","&Truncate",
                'c','h','t');
  switch (ret) {
    case 1:                                                  /* do nothing */
      return false;
      break;
    case 2:                                         /* fix embedded header */
      rblocks = blocks;
      write_header();
      get_header(str_fname);
      break;
    case 3:                                                /* truncate rom */
      blocks = rblocks;
      fsize = rblocks*16384;                    /* calculate new file size */
      if (smd) fsize+=512;
      truncate(str_fname,fsize);                          /* truncate file */
      file_info[file_idx]->size = fsize;  /* and change file size in array */
      if (smd) {                         /* fix smd header if there is one */
        f = fopen(str_fname,"rb+");
        if (f==NULL) {
          ok_alert(io_err,0,0);
          return false;
        }
        b = blocks;
        fputc(b,f);
        fclose(f);
      }
      get_header(str_fname);
      break;
  }
}

int diagnose_rom()
/* try to find out why the real rom size is different from the one */
/* in the embedded header */
{
  int i;
  int j;
  int equal;
  int diagnosis;
  int offset;
  word fill;
  block16k bufh;
  block16k buf1;
  block16k buf2;
  word buf3[8192];

  if (blocks > rblocks)
    diagnosis = long_dump;
  else
    diagnosis = short_dump; /* cannot be equal if this function was called */
  f = fopen(str_fname,"rb");
  if (f==NULL) {
    ok_alert(io_err,0,0);
    return false;
  }
  if (smd) {
    fseek(f,512,SEEK_SET);                              /* skip SMD header */
    offset = 512;
  }
  else offset =0;
  /* now try do make a diagnosis */
  if (blocks == rblocks << 1) {
    /* check for double dump */
    equal = true;
    for (i = 0; i < blocks >>1; i++) {
      fseek(f,(i << 14) + offset ,SEEK_SET);
      fread(buf1,1,16384,f);
      if (smd) {
        decode_block(buf1,bufh);
        memcpy(buf1,bufh,16384);
      }
      fseek(f,((i + (blocks >>1))<<14) + offset, SEEK_SET);
      fread(buf2,1,16384,f);
      if (smd) {
        decode_block(buf2,bufh);
        memcpy(buf2,bufh,16384);
      }
      for (j=0; j < 16384; j++) {
        if (buf1[j]!=buf2[j])
          if (i || j>0x1ff)      /* skip embedded header: i=0, j < 0x200 ) */
            equal = false;
      }
    }
    if (equal) diagnosis = double_dump;
  }
  else if (rblocks < blocks) {
    /* check for empty blocks */
    fseek (f, (rblocks << 14) + offset, SEEK_SET);   /* first surplus block*/
    fread(&fill,2,1,f);                                  /* get first word */
    fseek (f, (rblocks << 14) + offset, SEEK_SET);   /* first surplus block*/
    equal = true;
    for (i = 0; i < blocks-rblocks; i++) {
      fread(buf3,1,16384,f);
      for (j=0; j < 8192; j+=2) {
        if (buf3[j] != fill)
          equal = false;
      }
    }
    if (equal)
      diagnosis = empty_dump;
    else {
      /* check for intro */
      equal = false;
      fseek (f, (rblocks << 14) + offset, SEEK_SET);  /* 1st surplus block */
      for (i = 0; i < blocks-rblocks; i++) {
        fread(buf1,1,16384,f);
        if (smd) {
          decode_block(buf1,buf2);
          memcpy(buf1,buf2,16384);
        }
        for (j=0; j < 16383; j++) {
        if (buf1[j]=='b' || buf1[j]=='B')
           if (buf1[j+1]=='y' || buf1[j+1]=='Y')
              equal = true;
        }
      }
      if (equal) diagnosis = intro_dump;
    }
  }
  fclose(f);
  switch (diagnosis) {
    case double_dump:
      fix_rom("Size Mismatch: this rom is dumped twice for sure.",
              "I advise to truncate it.",
              "Fixing the header is not recommended.",
              diagnosis);
      break;
    case intro_dump:
      fix_rom("Size Mismatch: this rom has probably an intro or trainer.",
              "I advise to fix the header.",
              "Truncating will most probably destroy it.", diagnosis);
      break;
    case empty_dump:
      fix_rom("Size Mismatch: this rom has surplus empty pages.",
              "I advise to truncate it.",
              "Fixing the header is not recommended.", diagnosis);
      break;
    case long_dump:
      fix_rom("Size Mismatch: the rom is longer than reported in the header.",
              "Truncating it could destroy it",
              "while fixing only the header could waste space", diagnosis);
      break;
    case short_dump:
      fix_rom("Size Mismatch: the rom is shorter than reported in the header.",
              "The rom or its header could be corrupt but only the header",
              "can be fixed. Maybe you should look for another copy of this rom",
              diagnosis);
      break;
  }
}

void before_change()
/* called everytime the changes in header could get lost */
{
  if (changed)
    if (alert3d("Write Header","or discard changes",0,"&Write","&Discard",0,'w','d',0)==1)
      write_header();
  changed=bin=smd=blocks=rblocks=str_fname[0]=japan=usa=europe=0;
  strcpy(str_info,"<no rom>");
  clean_header();
}

void do_change(char p[])
/* called if path is changed */
{
  strcpy(str_path,p);
  before_change();
  change_path(str_path);
}

int main( int argc, char *argv[] )
{
  debug("MegaTool started");
  _crt0_startup_flags |= _CRT0_FLAG_NO_LFN;      /* disable long filenames */
  __djgpp_set_ctrl_c(0);                     /* treat ctrl-c as normal key */
  get_gmode();
  /* put together the path of the ini files and read cfg */
  strcpy(str_ini,argv[0]);
  clip_path(str_ini);
  conv_slash(str_ini);
  read_cfg();
  read_local_setup();
  /* do some initialization */
  init_allegro();
  strcpy(str_path,str_startpath);
  change_path(str_path);
  centre_dialog(about_dialog);
  centre_dialog(fr_dialog);
  centre_dialog(setup_dialog);
  centre_dialog(lsetup_dialog);
  centre_dialog(rename_dialog);
  /* main loop */
  while(!leave) {
    check_disabled();
    ret = do_dialog( my_dialog, find_dialog_focus(my_dialog));
    /* react on dialog event */
    switch( ret ) {
      case button_quit:
        before_change();
        leave=true; break;
      case button_about:
        popup_dialog(about_dialog, -1);
        break;
      case button_write:
        if (changed) write_header();
        break;
      case button_mlist:
        old_file_idx = file_idx;
        create_list();
        change_path(str_path);
        file_idx = old_file_idx;
        if (!file_info[file_idx]->dir)
          check_rom();
        else {
          clean_header();
          bin=smd=blocks=rblocks=changed=0;
        }
        break;
      case edit_overseas:
        changed=true;
        break;
      case edit_domestic:
        changed=true;
        break;
      case check_country_j:
        japan=!japan;
        changed=true;
        break;
      case check_country_u:
        usa=!usa;
        changed=true;
        break;
      case check_country_e:
        europe=!europe;
        changed=true;
        break;
      case button_a:
        do_change("A:");
        break;
      case button_b:
        do_change("B:");
        break;
      case button_c:
        do_change("C:");
        break;
      case button_d:
        do_change("D:");
        break;
      case button_e:
        do_change("E:");
        break;
      case button_f:
        do_change("F:");
        break;
      case button_g:
        strcpy(str_path,"G:");
        do_change("G:");
        break;
      case edit_path:
        before_change();
        change_path(str_path);
        break;
      case file_list:
        if ((file_idx==my_dialog[file_list].d1) && (bin || smd)) run_game();
        else {
          before_change();
          file_idx = my_dialog[file_list].d1;
          if (file_info[file_idx]->dir) {
            chdir(file_info[file_idx]->name);
            getcwd(str_path,127);
            conv_slash(str_path);
            get_files();
            my_dialog[file_list].d1 = 0; my_dialog[file_list].d2 = 0;
          }
          else {
            if (check_rom())
              if ( (blocks != rblocks) && (bin || smd) )
                diagnose_rom();
          }
        }
        break;
      case button_convert:
        convert();
        get_files();
        break;
      case button_calc_crc:
        calc_crc();
        break;
      case button_gsetup:
        global_setup();
        break;
      case button_lsetup:
        local_setup();
        break;
      case radio_all:
        strcpy(str_wild,wild_all);
        wild = w_all;
        get_files();
        break;
      case radio_rom:
        strcpy(str_wild,"*.SMD");
        wild = w_rom;
        get_files();
        break;
      case text_country:
        full_country();
        break;
      case button_rename:
        strcpy(str_exec, str_fname);          /* save filename in str_exec */
        strcpy(str_rename,str_fname);   /* init rename name with file name */
        ret = popup_dialog(rename_dialog, -1);
        if (ret==rename_button_rename) {
          before_change();                  /* str_fname is deleted here ! */
          doserror=_rename(str_exec,str_rename);
          get_files(true);
        }
        break;
      case button_delete:
        if (alert3d("Really delete file",str_fname,0,"&Yes","&No",0,'y','n',0)==1) {
          strcpy(str_exec, str_fname);        /* save filename in str_exec */
          before_change();                  /* str_fname is deleted here ! */
          doserror=remove(str_exec);
          get_files(true);
        }
        break;
      case button_join:
        join_rom(file_idx);
        get_files(true);
        break;
      case button_run:
        run_game();
    }
  }
  /* close down */
  change_path(str_ini); /* change back to the path MegaTool was started in */
  /*shutdown_gfx();*/   /* allegro's confused, so let's shut down manually */
  free_file_info();
  write_local_setup();
  for (i=0; i<opt_num; i++) kill_local_option(i);
  allegro_exit();
  set_gmode();
  exit(0);
}

